# -*- coding: utf-8 -*-

from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.resource import ResourceManagementClient
from chaoslib.types import Configuration, Secrets

from chaosazure import auth
from chaosazure.machine.picker import pick_machines

__all__ = ["describe_machines", "count_machines"]


def describe_machines(configuration: Configuration = None,
                      secrets: Secrets = None):
    """
    Describe Azure machines.
    """
    with auth(secrets) as cred:
        machines = pick_subscribed_machines(configuration, cred)
        return machines


def count_machines(configuration: Configuration = None,
                   secrets: Secrets = None) -> int:
    """
    Return count of Azure machines.
    """
    with auth(secrets) as cred:
        machines = pick_subscribed_machines(configuration, cred)
        return len(machines)


def pick_subscribed_machines(configuration, cred):
    azure_subscription_id = configuration['azure']['subscription_id']
    resource_client = ResourceManagementClient(cred, azure_subscription_id)
    compute_client = ComputeManagementClient(cred, azure_subscription_id)
    resource_groups_list = resource_client.resource_groups.list()
    machines_list_all = compute_client.virtual_machines.list_all()
    rg = configuration['azure']['resource_groups'].split(',')
    machines = pick_machines(resource_groups_list, machines_list_all, rg)
    return machines



# -*- coding: utf-8 -*-
from typing import Any, Dict, Generator

__all__ = ["ChaosParameters", "ServiceFabricAuth"]


ServiceFabricAuth = Generator[Dict[str, Any], None, None]
ChaosParameters = Dict[str, Any]



import random

from logzero import logger


def pick_machine_randomly(subscribed_resource_groups, subscribed_instances,
                          configured_resource_groups=None):
    machines = pick_machines(subscribed_resource_groups, subscribed_instances,
                             configured_resource_groups)
    return random.choice(machines)


def pick_machines(subscribed_resource_groups, subscribed_machines,
                  configured_resource_groups=None):
    machines = []

    if not configured_resource_groups or configured_resource_groups is None:
        logger.info("Client will pick all subscribed machines")
        for machine in subscribed_machines:
            machines.append(machine)
    else:
        for resource_group in subscribed_resource_groups:
            if resource_group.name in configured_resource_groups:
                for machine in subscribed_machines:
                    machine_id_lower = machine.id.lower()
                    group_id_lower = resource_group.id.lower()
                    if machine_id_lower.startswith(group_id_lower):
                        machines.append(machine)

    return machines



# -*- coding: utf-8 -*-
from typing import Any, Dict

from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.resource import ResourceManagementClient
from chaoslib.types import Configuration, Secrets
from logzero import logger

from chaosazure import auth
from chaosazure.machine.picker import pick_machine_randomly

__all__ = ["delete_machine", "poweroff_machine", "restart_machine"]


def delete_machine(configuration: Configuration = None,
                   secrets: Secrets = None) -> Dict[str, Any]:
    """
    Delete machine randomly.
    """
    logger.debug(
        'Starting delete_machine: configuration=[{}]'.format(configuration))

    with auth(secrets) as cred:
        compute_client, machine = __pick_chaos_machine(configuration, cred)

        resource_group_name = machine.id.split('/')[4].lower()
        logger.debug('Deletion chaos takes over for machine {}.{}'.format(
            resource_group_name, machine.name))
        async_restart = compute_client.virtual_machines.delete(
            resource_group_name, machine.name)
        result = async_restart.result()
        logger.debug(
            'Operation status for machine deletion: {}'.format(result.status))

        return result


def poweroff_machine(configuration: Configuration = None,
                     secrets: Secrets = None) -> Dict[str, Any]:
    """
    Power off machine randomly.
    """
    logger.debug(
        'Starting poweroff_machine: configuration=[{}]'.format(configuration))
    with auth(secrets) as cred:
        compute_client, machine = __pick_chaos_machine(configuration, cred)

        resource_group_name = machine.id.split('/')[4].lower()
        logger.debug('Powering off chaos takes over for machine {}.{}'.format(
            resource_group_name, machine.name))
        async_restart = compute_client.virtual_machines.power_off(
            resource_group_name, machine.name)
        result = async_restart.result()
        logger.debug(
            'Operation status for machine power off: {}'.format(result.status))

        return result


def restart_machine(configuration: Configuration = None,
                    secrets: Secrets = None) -> Dict[str, Any]:
    """
    Restart machine randomly.
    """
    logger.debug(
        'Starting restart_machine: configuration=[{}]'.format(configuration))
    with auth(secrets) as cred:
        compute_client, machine = __pick_chaos_machine(configuration, cred)

        resource_group_name = machine.id.split('/')[4].lower()
        logger.debug('Restarting chaos takes over for machine {}.{}'.format(
            resource_group_name, machine.name))
        async_restart = compute_client.virtual_machines.restart(
            resource_group_name, machine.name)
        result = async_restart.result()
        logger.debug(
            'Operation status for machine restart: {}'.format(result.status))

        return result


def __pick_chaos_machine(configuration, cred):
    azure_subscription_id = configuration['azure']['subscription_id']
    resource_client = ResourceManagementClient(cred, azure_subscription_id)
    compute_client = ComputeManagementClient(cred, azure_subscription_id)
    resource_groups_list = resource_client.resource_groups.list()
    machines_list_all = compute_client.virtual_machines.list_all()
    groups = configuration['azure']['resource_groups'].split(',')
    machine = pick_machine_randomly(resource_groups_list, machines_list_all,
                                    groups)
    return compute_client, machine



# -*- coding: utf-8 -*-
import json
from typing import Any, Dict

from logzero import logger
import requests

from chaoslib.exceptions import FailedActivity
from chaoslib.types import Configuration, Secrets

from chaosazure import sf_auth
from chaosazure.types import ChaosParameters

__all__ = ["start_chaos", "stop_chaos"]


def start_chaos(parameters: ChaosParameters, timeout: int = 60,
                configuration: Configuration = None,
                secrets: Secrets = None) -> Dict[str, Any]:
    """
    Start Chaos in your cluster using the given `parameters`. This is a mapping
    of keys as declared in the Service Fabric API:

    https://docs.microsoft.com/en-us/rest/api/servicefabric/sfclient-v60-model-chaosparameters

    Please see the :func:`chaosazure.fabric.auth` help for more information
    on authenticating with the service.
    """  # noqa: E501
    with sf_auth(configuration, secrets) as info:
        url = "{}/Tools/Chaos/$/Start".format(info["endpoint"])

        qs = {"api-version": "6.0"}
        if timeout is not None:
            qs["timeout"] = timeout

        r = requests.post(
            url, headers={"Accept": "application/json"},
            verify=info["verify"], params=qs, json=parameters)

        if r.status_code != 200:
            error = r.json()
            raise FailedActivity(
                "Service Fabric Chaos failed to start: {}".format(
                    json.dumps(error)))

        logger.debug("chaos started succesfully")

        return r.json()


def stop_chaos(timeout: int = 60, configuration: Configuration = None,
               secrets: Secrets = None) -> Dict[str, Any]:
    """
    Stop Chaos in your cluster.

    Please see the :func:`chaosazure.fabric.auth` help for more information
    on authenticating with the service.
    """
    with sf_auth(configuration, secrets) as info:
        url = "{}/Tools/Chaos/$/Stop".format(info["endpoint"])

        qs = {"api-version": "6.0"}
        if timeout is not None:
            qs["timeout"] = timeout

        r = requests.post(
            url, headers={"Accept": "application/json"}, verify=info["verify"],
            params=qs)

        if r.status_code != 200:
            error = r.json()
            raise FailedActivity(
                "Service Fabric Chaos failed to stop: {}".format(
                    json.dumps(error)))

        logger.debug("chaos stopped succesfully")

        return r.json()



# -*- coding: utf-8 -*-
from datetime import datetime, timezone
import json
from typing import Any, Dict

import dateparser
from logzero import logger
import requests

from chaoslib.exceptions import FailedActivity
from chaoslib.types import Configuration, Secrets

from chaosazure import sf_auth

__all__ = ["chaos_report"]


def chaos_report(timeout: int = 60, start_time_utc: str = None,
                 end_time_utc: str = None, configuration: Configuration = None,
                 secrets: Secrets = None) -> Dict[str, Any]:
    """
    Get Chaos report using following the Service Fabric API:

    https://docs.microsoft.com/en-us/rest/api/servicefabric/sfclient-v60-model-chaosparameters

    Please see the :func:`chaosazure.fabric.auth` help for more information
    on authenticating with the service.
    """  # noqa: E501
    with sf_auth(configuration, secrets) as info:
        url = "{}/Tools/Chaos/$/Report".format(info["endpoint"])

        qs = {"api-version": "6.0"}
        if timeout is not None:
            qs["timeout"] = timeout
        if start_time_utc is not None:
            qs["StartTimeUtc"] = datetime_to_ticks(start_time_utc)
        if end_time_utc is not None:
            qs["EndTimeUtc"] = datetime_to_ticks(end_time_utc)

        r = requests.get(
            url, headers={"Accept": "application/json"},
            verify=info["verify"], params=qs)

        if r.status_code != 200:
            error = r.json()
            raise FailedActivity(
                "Service Fabric Chaos failed to get report: {}".format(
                    json.dumps(error)))

        logger.debug("chaos report fetched succesfully")

        return r.json()


def datetime_to_ticks(when: str) -> int:
    """
    Computes the number of ticks until the relative time given.

    ```python
    >>> datetime_to_ticks("2 mns ago")
    ```
    """
    dt = dateparser.parse(when, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    if not dt:
        raise FailedActivity("failed parsing moment: {}".format(when))

    span = dt - datetime(1, 1, 1, tzinfo=timezone.utc)
    return int(span.total_seconds() * 10**7)



# -*- coding: utf-8 -*-

"""Top-level package for chaostoolkit-azure."""
import configparser
import contextlib
import os
import tempfile
from typing import List

from chaoslib.discovery import initialize_discovery_result, discover_actions, \
    discover_probes
from chaoslib.exceptions import FailedActivity
from chaoslib.types import Discovery, DiscoveredActivities, Secrets, \
    Configuration
from logzero import logger
from msrestazure.azure_active_directory import ServicePrincipalCredentials

from chaosazure.types import ServiceFabricAuth

__all__ = ["sf_auth", "auth", "discover", "__version__"]
__version__ = '0.2.0'


@contextlib.contextmanager
def sf_auth(configuration: Configuration,
            secrets: Secrets) -> ServiceFabricAuth:
    """
    Attempt to load the Service Fabric authentication information from a local
    configuration file or the passed `configuration` mapping. The latter takes
    precedence over the local configuration file.

    If you provide a configuration and secrets dictionary, the returned mapping
    will be created from their content. For instance, you could have:

    Configuration mapping (in your experiment file):
    ```python
    {
        "endpoint": "https://XYZ.westus.cloudapp.azure.com:19080",
        "verify_tls": False,
        "use_ca": False
    }
    ```

    Secrets mapping (in your experiment file):
    ```python
    {
        "azure": {
            "security": "pem",
            "pem_content": {
                "type": "env",
                "key": "AZURE_SERVICE_FABRIC_PEM"
            }
        }
    }
    ```

    In that case, the PEM content will be read from the local environment
    variable `AZURE_SERVICE_FABRIC_PEM` that you will have populated before
    hand. The content will be saved by the extension into a temporary file
    before being used to authenticate.

    You could also simply have that file ready instead:

    Secrets mapping (in your experiment file):
    ```python
    {
        "azure": {
            "security": "pem",
            "pem_path": "./party-cluster-XYZ-client-cert.pem"
        }
    }
    ```

    If you want to load the information from a local Service Fabric
    config file, set the `config_path` key in the `configuration mapping.

    Configuration mapping (in your experiment file):
    ```python
    {
        "config_path": "~/.sfctl/config"
    }
    ```
    The path will be expanded.

    The authentification file should look like this:

    ```ini
    [servicefabric]
    endpoint = https://XYZ.westus.cloudapp.azure.com:19080
    no_verify = true
    use_ca = false
    security = pem
    pem_path = ./party-cluster-XYZ-client-cert.pem
    ```

    No matter the input, the yielded dictionary looks like this:

    ```python
    {
        "endpoint": "https://XYZ.westus.cloudapp.azure.com:19080",
        "verify": False,
        "security": {
            "type": "pem",
            "path": "./party-cluster-XYZ-client-cert.pem"
        }
    }
    ```

    Using this function goes as follows:

    ```python
    with auth(configuration, secrets) as info:
        url = "{}{}".format(
            info["endpoint"], "/Tools/Chaos/$/Start?api-version=6.0")

        r = requests.get(
            url, cert=info["security"]["path"], verify=info["verify"])

    """
    c = configuration or {}
    s = secrets or {}

    config_path = c.get("config_path")
    endpoint = c.get("endpoint", s.get("endpoint"))

    if config_path:
        config_path = os.path.expanduser(config_path)
        if not os.path.exists(config_path):
            raise FailedActivity(
                "Service Fabric configuration file not found at {}".format(
                    config_path
                ))

        with open(config_path) as f:
            parser = configparser.ConfigParser()
            parser.read_file(f)

            pem_path = parser.get("servicefabric", "pem_path")
            if not pem_path:
                raise FailedActivity("cannot find {}".format(pem_path))

            yield {
                "endpoint": parser.get("servicefabric", "endpoint"),
                "verify": not (
                        parser.get("servicefabric", "no_verify") != "true"),
                "security": {
                    "type": parser.get("servicefabric", "security"),
                    "path": pem_path
                }
            }

    elif endpoint:
        verify_tls = c.get("verify_tls", s.get("verify_tls", True))
        use_ca = c.get("use_ca", s.get("use_ca", True))
        security_kind = s.get("security", c.get("security", "pem"))
        pem_path = s.get("pem_path", c.get("pem_path", None))
        pem_content = s.get("pem_content", c.get("pem_content", None))

        info = {
            "endpoint": endpoint,
            "verify": verify_tls,
            "security": {
                "type": security_kind,
                "path": pem_path
            }
        }

        if not pem_path or (not os.path.exists(pem_path) and pem_content):
            # the file will be deleted when we leave the context block
            with tempfile.NamedTemporaryFile(mode="w+",
                                             encoding='utf-8') as pem_path:

                pem_path.write(pem_content)
                pem_path.seek(0)

                info["security"]["pem_path"] = pem_path.name
                yield info
        else:
            yield info
    else:
        raise FailedActivity(
            "Service Fabric client needs to know how to authenticate")


@contextlib.contextmanager
def auth(secrets: Secrets) -> ServicePrincipalCredentials:
    """
    Attempt to load the Azure authentication information from a local
    configuration file or the passed `configuration` mapping. The latter takes
    precedence over the local configuration file.

    If you provide a secrets dictionary, the returned mapping will
    be created from their content. For instance, you could have:

    Secrets mapping (in your experiment file):
    ```json
    {
        "azure": {
            "client_id": "AZURE_CLIENT_ID",
            "client_secret": "AZURE_CLIENT_SECRET",
            "tenant_id": "AZURE_TENANT_ID"
        }
    }
    ```

    The client_id, tenant_id, and client_secret content will be read
    from the specified local environment variables, e.g. `AZURE_CLIENT_ID`,
    `AZURE_TENANT_ID`, and `AZURE_CLIENT_SECRET` that you will have populated
    before hand.
    ```

    Using this function goes as follows:

    ```python
    with auth(secrets) as cred:
        azure_subscription_id = configuration['azure']['subscription_id']
        resource_client = ResourceManagementClient(cred, azure_subscription_id)
        compute_client = ComputeManagementClient(cred, azure_subscription_id)
    ```
    """
    azure_secrets = secrets or {}

    if not azure_secrets:
        raise FailedActivity("Azure secrets not found")

    client_id, client_secret, tenant_id = __fetch_secrets(azure_secrets)
    __check_secrets(client_id, client_secret, tenant_id)
    credentials = __generate_service_principal_credentials(client_id,
                                                           client_secret,
                                                           tenant_id)

    yield credentials


def discover(discover_system: bool = True) -> Discovery:
    """
    Discover Azure capabilities offered by this extension.
    """
    logger.info("Discovering capabilities from chaostoolkit-azure")

    discovery = initialize_discovery_result(
        "chaostoolkit-azure", __version__, "azure")
    discovery["activities"].extend(__load_exported_activities())
    return discovery


###############################################################################
# Private functions
###############################################################################
def __generate_service_principal_credentials(client_id, client_secret,
                                             tenant_id):
    credentials = ServicePrincipalCredentials(
        client_id=client_id,
        secret=client_secret,
        tenant=tenant_id
    )
    return credentials


def __check_secrets(client_id, client_secret, tenant_id):
    if not client_id or not client_secret or not tenant_id:
        raise FailedActivity("Client could not find Azure credentials")


def __fetch_secrets(azure_secrets):
    #
    # fetch secrets from environment
    #
    client_id = os.getenv(azure_secrets.get('client_id'))
    client_secret = os.getenv(azure_secrets.get('client_secret'))
    tenant_id = os.getenv(azure_secrets.get('tenant_id'))

    #
    # fetch secrets from file content itself when empty
    #
    if not client_id:
        client_id = azure_secrets.get('client_id')
    if not client_secret:
        client_secret = azure_secrets.get('client_secret')
    if not tenant_id:
        tenant_id = azure_secrets.get('tenant_id')
    return client_id, client_secret, tenant_id


def __load_exported_activities() -> List[DiscoveredActivities]:
    """
    Extract metadata from actions and probes exposed by this extension.
    """
    activities = []
    activities.extend(discover_actions("chaosazure.fabric.actions"))
    activities.extend(discover_probes("chaosazure.fabric.probes"))
    activities.extend(discover_actions("chaosazure.machine.actions"))
    activities.extend(discover_probes("chaosazure.machine.probes"))
    return activities



#!/usr/bin/env python
"""chaostoolkit-azure extension builder and installer"""

import sys
import io

import setuptools

sys.path.insert(0, ".")
from chaosazure import __version__
sys.path.remove(".")

name = 'chaostoolkit-azure'
desc = 'Chaos Toolkit Extension for Microsoft Azure'

with io.open('README.md', encoding='utf-8') as strm:
    long_desc = strm.read()

classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'License :: Freely Distributable',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: Implementation',
    'Programming Language :: Python :: Implementation :: CPython'
]
author = "chaostoolkit Team"
author_email = 'contact@chaostoolkit.org'
url = 'http://chaostoolkit.org'
license = 'Apache License Version 2.0'
packages = [
    'chaosazure',
    'chaosazure.fabric',
    'chaosazure.machine'
]

needs_pytest = set(['pytest', 'test']).intersection(sys.argv)
pytest_runner = ['pytest_runner'] if needs_pytest else []

test_require = []
with io.open('requirements-dev.txt') as f:
    test_require = [l.strip() for l in f if not l.startswith('#')]

install_require = []
with io.open('requirements.txt') as f:
    install_require = [l.strip() for l in f if not l.startswith('#')]

setup_params = dict(
    name=name,
    version=__version__,
    description=desc,
    long_description=long_desc,
    classifiers=classifiers,
    author=author,
    author_email=author_email,
    url=url,
    license=license,
    packages=packages,
    include_package_data=True,
    install_requires=install_require,
    tests_require=test_require,
    setup_requires=pytest_runner,
    python_requires='>=3.5.*'
)


def main():
    """Package installation entry point."""
    setuptools.setup(**setup_params)


if __name__ == '__main__':
    main()
